@@ -144,14 +144,17 @@ typedef struct
   gpointer user_data;
 
   GtkWidget *window;
+  GtkEventController *window_close_controller;
   gboolean force_closing;
 } AdwDialogPrivate;
 
 static void adw_dialog_buildable_init (GtkBuildableIface *iface);
+static void adw_dialog_shortcut_manager_init (GtkShortcutManagerInterface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (AdwDialog, adw_dialog, GTK_TYPE_WIDGET,
                          G_ADD_PRIVATE (AdwDialog)
-                         G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE, adw_dialog_buildable_init))
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE, adw_dialog_buildable_init)
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_SHORTCUT_MANAGER, adw_dialog_shortcut_manager_init))
 
 static GtkBuildableIface *parent_buildable_iface;
 
@@ -180,6 +183,12 @@ enum {
 
 static guint signals[SIGNAL_LAST_SIGNAL];
 
+static void
+child_breakpoint_bin_notify_current_breakpoint_cb (AdwDialog *self)
+{
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CURRENT_BREAKPOINT]);
+}
+
 static gboolean
 map_tick_cb (AdwDialog *self)
 {
@@ -603,7 +612,6 @@ present_as_window (AdwDialog *self,
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
   GtkWidget *titlebar;
-  GtkEventController *shortcut_controller;
   GtkShortcut *shortcut;
 
   if (priv->window) {
@@ -622,9 +630,10 @@ present_as_window (AdwDialog *self,
   shortcut = gtk_shortcut_new (gtk_keyval_trigger_new (GDK_KEY_Escape, 0),
                                gtk_callback_action_new ((GtkShortcutFunc) maybe_close_cb, self, NULL));
 
-  shortcut_controller = gtk_shortcut_controller_new ();
-  gtk_shortcut_controller_add_shortcut (GTK_SHORTCUT_CONTROLLER (shortcut_controller), shortcut);
-  gtk_widget_add_controller (priv->window, shortcut_controller);
+  priv->window_close_controller = gtk_shortcut_controller_new ();
+  gtk_shortcut_controller_set_scope (GTK_SHORTCUT_CONTROLLER (priv->window_close_controller), GTK_SHORTCUT_SCOPE_MANAGED);
+  gtk_shortcut_controller_add_shortcut (GTK_SHORTCUT_CONTROLLER (priv->window_close_controller), shortcut);
+  gtk_widget_add_controller (GTK_WIDGET (self), priv->window_close_controller);
 
   if (parent) {
     GtkRoot *root = gtk_widget_get_root (parent);
@@ -776,6 +785,23 @@ ensure_focus (AdwDialog *self)
   return TRUE;
 }
 
+static gboolean
+toggle_adaptive_preview_cb (AdwDialog *self)
+{
+  GtkRoot *root;
+  gboolean open;
+
+  if (!adw_get_inspector_keybinding_enabled ())
+    return GDK_EVENT_PROPAGATE;
+
+  root = gtk_widget_get_root (GTK_WIDGET (self));
+
+  g_object_get (root, "adaptive-preview", &open, NULL);
+  g_object_set (root, "adaptive-preview", !open, NULL);
+
+  return GDK_EVENT_STOP;
+}
+
 static void
 adw_dialog_root (GtkWidget *widget)
 {
@@ -1005,6 +1031,16 @@ adw_dialog_set_property (GObject      *object,
   }
 }
 
+static void
+adw_dialog_real_closed (AdwDialog *dialog)
+{
+}
+
+static void
+adw_dialog_real_close_attempt (AdwDialog *dialog)
+{
+}
+
 static void
 adw_dialog_class_init (AdwDialogClass *klass)
 {
@@ -1024,6 +1060,9 @@ adw_dialog_class_init (AdwDialogClass *klass)
   widget_class->contains = adw_widget_contains_passthrough;
   widget_class->compute_expand = adw_widget_compute_expand;
 
+  klass->closed = adw_dialog_real_closed;
+  klass->close_attempt = adw_dialog_real_close_attempt;
+
   /**
    * AdwDialog:child:
    *
@@ -1251,6 +1290,8 @@ adw_dialog_class_init (AdwDialogClass *klass)
                                 (GtkShortcutFunc) open_inspector_cb, "b", FALSE);
   gtk_widget_class_add_binding (widget_class, GDK_KEY_D, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
                                 (GtkShortcutFunc) open_inspector_cb, "b", TRUE);
+  gtk_widget_class_add_binding (widget_class, GDK_KEY_M, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
+                                (GtkShortcutFunc) toggle_adaptive_preview_cb, NULL);
 
   gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
   gtk_widget_class_set_css_name (widget_class, "dialog");
@@ -1270,6 +1311,8 @@ adw_dialog_init (AdwDialog *self)
   priv->follows_content_size = FALSE;
   priv->presentation_mode = ADW_DIALOG_AUTO;
 
+  gtk_widget_set_limit_events (GTK_WIDGET (self), TRUE);
+
   priv->child_breakpoint_bin = adw_breakpoint_bin_new ();
   gtk_widget_set_overflow (priv->child_breakpoint_bin, GTK_OVERFLOW_VISIBLE);
   adw_breakpoint_bin_set_warning_widget (ADW_BREAKPOINT_BIN (priv->child_breakpoint_bin),
@@ -1280,6 +1323,10 @@ adw_dialog_init (AdwDialog *self)
   g_object_bind_property (self, "height-request",
                           priv->child_breakpoint_bin, "height-request",
                           G_BINDING_DEFAULT);
+  g_signal_connect_swapped (priv->child_breakpoint_bin,
+                            "notify::current-breakpoint",
+                            G_CALLBACK (child_breakpoint_bin_notify_current_breakpoint_cb),
+                            self);
 }
 
 static void
@@ -1305,6 +1352,11 @@ adw_dialog_buildable_init (GtkBuildableIface *iface)
   iface->add_child = adw_dialog_buildable_add_child;
 }
 
+static void
+adw_dialog_shortcut_manager_init (GtkShortcutManagerInterface *iface)
+{
+}
+
 /**
  * adw_dialog_new:
  *
@@ -1360,14 +1412,14 @@ adw_dialog_set_child (AdwDialog *self,
   g_return_if_fail (ADW_IS_DIALOG (self));
   g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
 
-  if (child)
-    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
-
   priv = adw_dialog_get_instance_private (self);
 
   if (priv->child == child)
     return;
 
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
+
   priv->child = child;
 
   adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->child_breakpoint_bin), child);
@@ -1902,6 +1954,9 @@ adw_dialog_close (AdwDialog *self)
     GtkWidget *window = priv->window;
     priv->window = NULL;
 
+    gtk_widget_remove_controller (GTK_WIDGET (self), priv->window_close_controller);
+    priv->window_close_controller = NULL;
+
     if (priv->closing_callback)
       priv->closing_callback (self, priv->user_data);
 
